import { world, system, ItemStack, BlockPermutation } from '@minecraft/server';
import { randomTickComponent, OnTickComponent, leafDrop } from "./OnTick"
import { treeInteractComponent, leafDefaultComponent } from './tree';
import { FenceInteractComponent, PlaceComponent, PlaceGateComponent, GateInteract } from './fence';
import { slabComponent } from './slabs';
import "./ybiomes_items.js";
import "./ybiomes_foods.js";
import 'door'
import 'trapdoor'

function main() {
  ItemClass.init()
}

// Add more mappings here as needed - Note, the doors need to be set on door.js at line 237.
const isSlab = new Set([
  "biomes:maple_slab",
  "biomes:rosewood_slab"
]);
const fenceSet = new Set([
  "biomes:maple_fence",
  "biomes:maple_fence_gate",
  "biomes:rosewood_fence",
  "biomes:rosewood_fence_gate"
]);
const fenceMap = new Map([
  ["biomes:rosewood_fence_inventory", "biomes:rosewood_fence"],
  ["biomes:fence_inventory", "biomes:maple_fence"]
]);
const leafSet = new Set([
  "biomes:maple_leaves_orange",
  "biomes:maple_leaves_red",
  "biomes:maple_leaves_yellow",
  "biomes:maple_leaves_brown",
  "biomes:orchard_leaves",
  "biomes:dark_cherry_leaves",
  "biomes:snowy_leaves",
  "biomes:walnut_leaves",
  "biomes:rosewood_leaves",
  "biomes:clear_spruce_leaves",
  "biomes:dried_spruce_leaves"
]);
const fruitSet = new Set([
  "biomes:orchard_leaves"
]);
const leafDropMap = new Map([
  ["biomes:maple_leaves", "biomes:maple_sapling"],
  ["biomes:rosewood_leaves", "biomes:rosewood_sapling"]
]);
const saplingBlocks = [
  {
    blockID: "biomes:maple_sapling",
    maxStage: 2,
    growChance: {
      numerator: 1,
      denominator: 3
    }
  },
  {
    blockID: "biomes:rosewood_sapling",
    maxStage: 2,
    growChance: {
      numerator: 1,
      denominator: 3
    }
  },
  {
    blockID: "biomes:orchard_leaves",
    maxStage: 3,
    growChance: {
      numerator: 1,
      denominator: 7 // adjust this to adjust speed of fruit growth
    }
  }
  // Add other sapling types here
];
const logSet = new Set([
  "biomes:maple_log",
  "biomes:rosewood_log"
]);
const sapSet = new Set([
  "biomes:maple_sapling",
  "biomes:rosewood_sapling"
]);
const logMap = new Map([
  ["biomes:maple_log", "biomes:stripped_maple_log"],
  ["biomes:rosewood_log", "biomes:stripped_rosewood_log"]
]);
const trapSet = new Set([
  "biomes:maple_trapdoor",
  "biomes:rosewood_trapdoor"
]);

// This handles the durality damage for all blocks (except slabs - handled in slabs.js) and removing the block above for fences and gates
const DestroyComponent = {
  onPlayerDestroy(e) {
    const { block, player } = e;
    const aboveBlock = block.above();
    const blockBelow = block.below();

    // Remove wiki:fence_inventory or invisible fence gate on top if present
    if ((aboveBlock.typeId.includes("_inventory")) || (fenceSet.has(aboveBlock.typeId) && aboveBlock.permutation.getState('ybrothers:invisible'))) {
      aboveBlock.setType("minecraft:air");
    }

    // add wiki:fence_inventory if fence present below
    if (fenceSet.has(blockBelow.typeId)) {
      block.setPermutation(
        BlockPermutation.resolve("ybrothers:fence_inventory", { "ybrothers:post": 1 })
      );
    }
    const inventory = player.getComponent("inventory");
    const selectedItem = inventory.container.getItem(player.selectedSlotIndex);
    if (
      selectedItem &&
      (selectedItem.typeId.includes("shovel") ||
        selectedItem.typeId.includes("hoe") ||
        selectedItem.typeId.includes("_axe") ||
        selectedItem.typeId.includes("_pickaxe") ||
        selectedItem.typeId.includes("shears"))
    ) {
      system.runTimeout(() => {
        applyDurabilityDamage(
          player,
          selectedItem,
          inventory,
          player.selectedSlotIndex
        );
      }, 1);
    }
  },
};

const customComponents = {
  "ybrothers:fence_interact": FenceInteractComponent,
  "ybrothers:on_player_destroy": DestroyComponent,
  "ybrothers:on_tick": OnTickComponent,
  "ybrothers:on_player_placed": PlaceComponent,
  "ybrothers:random_tick": randomTickComponent,
  "ybrothers:slab_interact": slabComponent,
  "ybrothers:interact_tree": treeInteractComponent,
  "ybrothers:gate_interact": GateInteract,
  "ybrothers:gate_placed": PlaceGateComponent,
  "ybrothers:leaf_place": leafDefaultComponent
};

let lastBrokenBlockType = null;
let lastBrokenBlockIsDouble = false;

// Use this world initialization event to register all custom components
world.beforeEvents.worldInitialize.subscribe(({ blockComponentRegistry }) => {
  for (const [componentName, componentImplementation] of Object.entries(customComponents)) {
    blockComponentRegistry.registerCustomComponent(componentName, componentImplementation);
  }
});

world.beforeEvents.playerBreakBlock.subscribe((eventData) => {
  const { block, player, dimension } = eventData;
  const blockBelow = block.below();
  const inventory = player.getComponent("inventory");
  const selectedItem = inventory.container.getItem(player.selectedSlotIndex);
  const { hasSilkTouch } = getRelevantEnchantments(selectedItem);

  if (selectedItem && leafSet.has(block.typeId) && selectedItem.typeId === "minecraft:shears" && !hasSilkTouch) {
    const spawnItem = block.typeId;
    system.runTimeout(() => {
      block.dimension.spawnItem(new ItemStack(spawnItem, 1), block.location);
    }, 1);
  } else if (leafSet.has(block.typeId) && !hasSilkTouch) {
    leafDrop(block, true);
  }

  if (isSlab.has(block.typeId)) {
    lastBrokenBlockType = block.typeId;
    lastBrokenBlockIsDouble =
      block.permutation.getState("ybrothers:double") === true;
    system.runTimeout(() => {
      if (!selectedItem) {
        slabDrop(block, dimension, player);

      } else if (
        selectedItem.typeId.includes("shovel") ||
        selectedItem.typeId.includes("hoe") ||
        selectedItem.typeId.includes("_axe") ||
        selectedItem.typeId.includes("_pickaxe")
      ) {
        applyDurabilityDamage(
          player,
          selectedItem,
          inventory,
          player.selectedSlotIndex
        );
        slabDrop(block, dimension, player);
      } else {
        slabDrop(block, dimension, player);
      }
    }, 1);
  } else if (fenceSet.has(blockBelow.typeId)) {
    system.runTimeout(() => {
      if (blockBelow.typeId.includes("ybrothers:") && blockBelow.typeId.includes("_gate")) {
        const currentStates = blockBelow.permutation.getAllStates();

        const cardinalDirection = currentStates['minecraft:cardinal_direction'];
        const newAboveStates = { ...currentStates };
        if (cardinalDirection === 'south') {
          delete newAboveStates['minecraft:cardinal_direction'];
        }
        newAboveStates['ybrothers:invisible'] = true;
        block.setPermutation(BlockPermutation.resolve(blockBelow.typeId, newAboveStates));

      } else if (blockBelow.typeId.includes("ybrothers:") && !blockBelow.typeId.includes("_gate")) {
        block.setPermutation(
          BlockPermutation.resolve("ybrothers:fence_inventory", { "ybrothers:post": 1 })
        );
      }
    }, 1);
  }
});

function slabDrop(block, dimension, player) {
  if (player.getGameMode() !== "creative") {
    if (lastBrokenBlockType) {
      const blockData = lastBrokenBlockType;

      if (blockData) {
        let dropItem;

        if (
          lastBrokenBlockIsDouble
        ) {
          dropItem = new ItemStack(lastBrokenBlockType, 1);
        }
        if (dropItem) {
          dimension.spawnItem(dropItem, block.location);
        }
      }
      lastBrokenBlockType = null;
      lastBrokenBlockIsDouble = false;
    }
  }
}

function applyDurabilityDamage(player, item, inventory, slotIndex) {
  const durabilityComponent = item.getComponent("minecraft:durability");
  if (durabilityComponent) {
    const { unbreakingLevel } = getRelevantEnchantments(item);

    if (Math.random() < 1 / (unbreakingLevel + 1)) {
      const newDamage = durabilityComponent.damage + 1;
      if (newDamage >= durabilityComponent.maxDurability) {
        inventory.container.setItem(slotIndex, undefined);
        player.playSound("random.break");
      } else {
        durabilityComponent.damage = newDamage;
        inventory.container.setItem(slotIndex, item);
      }
    }
  }
}

function getRelevantEnchantments(item) {
  let unbreakingLevel = 0;
  let hasSilkTouch = false;

  try {
    const enchantableComponent = item.getComponent("minecraft:enchantable");
    if (enchantableComponent) {
      const enchantments = enchantableComponent.getEnchantments();
      for (const enchant of enchantments) {
        if (enchant.type.id === "unbreaking") {
          unbreakingLevel = enchant.level;
        } else if (enchant.type.id === "silk_touch") {
          hasSilkTouch = true;
        }
      }
    }
  } catch (error) {
  }
  return { unbreakingLevel, hasSilkTouch };
}



export {
  applyDurabilityDamage,
  getRelevantEnchantments,
  isSlab,
  fenceSet,
  fenceMap,
  leafSet,
  leafDropMap,
  saplingBlocks,
  logSet,
  sapSet,
  logMap,
  trapSet,
  fruitSet
}

system.run(main());